/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: JORAMHelper.java 2438 2010-02-24 16:20:30Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server.impl.jms;

import java.net.ConnectException;
import java.util.Arrays;
import java.util.List;

import javax.jms.Topic;

import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.joram.client.jms.admin.AdminModule;
import org.objectweb.joram.client.jms.admin.Server;
import org.ow2.cmi.controller.server.ClusterView;
import org.ow2.cmi.controller.server.Domain;
import org.ow2.cmi.controller.server.ServerView;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Helper to create topics on a JORAM server.
 * @author Loris Bouzonnet
 */
public final class JORAMHelper {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(JORAMHelper.class);


    /**
     * Helper class: no public constructor.
     */
    private JORAMHelper() {
        // Do nothing
    }

    public static Topic createTopic(
            final String topicName, final Topic extTopic,
            final boolean clustered, final boolean hierarchical) throws JORAMServerException {
        org.objectweb.joram.client.jms.Topic topic;
        try {
            topic = org.objectweb.joram.client.jms.Topic.create(topicName);
            topic.setFreeReading();
            topic.setFreeWriting();
            if(hierarchical && extTopic instanceof org.objectweb.joram.client.jms.Topic) {
                topic.setParent(
                        (org.objectweb.joram.client.jms.Topic) extTopic);
            } else if(clustered && extTopic instanceof org.objectweb.joram.client.jms.Topic) {
                topic.addClusteredTopic(
                        (org.objectweb.joram.client.jms.Topic) extTopic);
            }
        } catch (ConnectException e) {
            logger.error("Unable to create the topic with name " + topicName, e);
            throw new JORAMServerException("Unable to create the topic with name " + topicName, e);
        } catch (AdminException e) {
            logger.error("Unable to create the topic with name " + topicName, e);
            throw new JORAMServerException("Unable to create the topic with name " + topicName, e);
        }
        return topic;
    }

    public static void checkMyTopic(final Topic localTopic, final Topic extTopic,
            final boolean clustered, final boolean hierarchical) throws JORAMServerException {
        if(hierarchical
                && localTopic instanceof org.objectweb.joram.client.jms.Topic
                && extTopic instanceof org.objectweb.joram.client.jms.Topic) {
            try {
                Topic topic =
                    ((org.objectweb.joram.client.jms.Topic) localTopic).getHierarchicalFather();
                if(topic != null) {
                    if(!topic.equals(extTopic)) {
                        logger.error("Not a hierarchical topic as expected.");
                        throw new JORAMServerException("Not a hierarchical topic as expected.");
                    }
                } else {
                    ((org.objectweb.joram.client.jms.Topic) localTopic).setParent(
                            (org.objectweb.joram.client.jms.Topic) extTopic);
                }
            } catch (ConnectException e) {
                logger.error("Unable to check the use of hierarchical topic", e);
                throw new JORAMServerException("Unable to check the use of hierarchical topic", e);
            } catch (AdminException e) {
                logger.error("Unable to check the use of hierarchical topic", e);
                throw new JORAMServerException("Unable to check the use of hierarchical topic", e);
            }
        } else if(clustered
                && localTopic instanceof org.objectweb.joram.client.jms.Topic
                && extTopic instanceof org.objectweb.joram.client.jms.Topic) {
            try {
                List<?> fellows =
                    ((org.objectweb.joram.client.jms.Topic) localTopic).getClusterFellows();
                if(!fellows.contains(localTopic)) {
                    ((org.objectweb.joram.client.jms.Topic) localTopic).addClusteredTopic(
                            (org.objectweb.joram.client.jms.Topic) extTopic);
                }
            } catch (ConnectException e) {
                logger.error("Unable to check the use of clustered topic", e);
                throw new JORAMServerException("Unable to check the use of clustered topic", e);
            } catch (AdminException e) {
                logger.error("Unable to check the use of clustered topic", e);
                throw new JORAMServerException("Unable to check the use of clustered topic", e);
            }
        }
    }

    /**
     * @param config
     * @throws JORAMServerException
     */
    public static void setDistributedConfig(final Config config, final ClusterView clusterView) throws JORAMServerException {
        String domainName = config.getDomainName();
        if(domainName == null) {
            logger.error("Unable to set a distributed config without domain name!");
            throw new JORAMServerException("Unable to set a distributed config without domain name!");
        }

        if (logger.isDebugEnabled()) {
            try {
               logger.debug("Servers belonging to the domain {0}: {1}", domainName, AdminModule.getServersIds(domainName));
            } catch (ConnectException e) {
                logger.debug("Unable to retrieve the servers belonging to the domain " + domainName, e);
            } catch (AdminException e) {
                logger.debug("Unable to retrieve the servers belonging to the domain " + domainName, e);
            }
        }

        try {
            addDomain(config);
            if (clusterView != null) {
                for (ServerView serverView : clusterView.getServerViews(domainName)) {
                    addServer(serverView, domainName);
                }
            }
        } catch (Exception e) {
            logger.error("Unable to administrate the JORAM server", e);
            throw new JORAMServerException("Unable to administrate the JORAM server", e);
        }
    }

    /**
     * @return
     * @throws JORAMServerException
     */
    public static String getServerName() throws JORAMServerException {
        try {
            return AdminModule.getLocalServer().getName();
        } catch (ConnectException e) {
            logger.error("Unable to retrieve the local server name", e);
            throw new JORAMServerException("Unable to retrieve the local server name", e);
        } catch (AdminException e) {
            logger.error("Unable to retrieve the local server name", e);
            throw new JORAMServerException("Unable to retrieve the local server name", e);
        }
    }

    /**
     * @param serverView
     * @param domainName
     * @throws JORAMServerException
     */
    public static void addServer(final ServerView serverView, final String domainName) throws JORAMServerException {

        // Update the server configuration
        Domain domain = serverView.getDomain(domainName);
        if (domain != null) {
            try {
                Server[] servers = AdminModule.getServers(domainName);
                for (Server server : servers) {
                    if (server.getId() == serverView.getId()) {
                        return;
                    }
                }
                AdminModule.addServer(
                        serverView.getId(), serverView.getServerRef().getHostname(),
                        domainName, domain.getPort(), serverView.getServerName());
                logger.debug("New server added to the domain {0}: {1}", serverView);
            } catch (ConnectException e) {
                logger.error("Unable to add the server " + serverView, e);
                throw new JORAMServerException("Unable to add the server " + serverView, e);
            } catch (AdminException e) {
                logger.error("Unable to add the server " + serverView, e);
                throw new JORAMServerException("Unable to add the server " + serverView, e);
            }
        }
    }

    /**
     * @param sid
     */
    public static void removeServer(final Topic topic, final short sid) throws JORAMServerException {

        if (topic != null && topic instanceof org.objectweb.joram.client.jms.Topic) {
            Topic superTopic;
            try {
                superTopic =
                    ((org.objectweb.joram.client.jms.Topic) topic).getHierarchicalFather();
            } catch (ConnectException e) {
                logger.error("Unable to check the use of hierarchical topic", e);
                throw new JORAMServerException("Unable to check the use of hierarchical topic", e);
            } catch (AdminException e) {
                logger.error("Unable to check the use of hierarchical topic", e);
                throw new JORAMServerException("Unable to check the use of hierarchical topic", e);
            }
            if (superTopic != null) {
                try {
                    ((org.objectweb.joram.client.jms.Topic) topic).unsetParent();
                } catch (ConnectException e) {
                    logger.error("Unable to unset the parent", e);
                    throw new JORAMServerException("Unable to unset the parent", e);
                } catch (AdminException e) {
                    logger.error("Unable to unset the parent", e);
                    throw new JORAMServerException("Unable to unset the parent", e);
                }
            } else {
                List<?> fellows;
                try {
                    fellows =
                        ((org.objectweb.joram.client.jms.Topic) topic).getClusterFellows();
                } catch (ConnectException e) {
                    logger.error("Unable to check the use of clustered topic", e);
                    throw new JORAMServerException("Unable to check the use of clustered topic", e);
                } catch (AdminException e) {
                    logger.error("Unable to check the use of clustered topic", e);
                    throw new JORAMServerException("Unable to check the use of clustered topic", e);
                }
                if (fellows != null && !fellows.isEmpty()) {
                    try {
                        ((org.objectweb.joram.client.jms.Topic) topic).removeFromCluster();
                    } catch (ConnectException e) {
                        logger.error("Unable to remove from the cluster", e);
                        throw new JORAMServerException("Unable to remove from the cluster", e);
                    } catch (AdminException e) {
                        logger.error("Unable to remove from the cluster", e);
                        throw new JORAMServerException("Unable to remove from the cluster", e);
                    }
                }
            }
        }

        // Update the server configuration
        try {
            AdminModule.removeServer(sid);
        } catch (ConnectException e) {
            logger.error("Unable to remove the server " + sid, e);
            throw new JORAMServerException("Unable to remove the server " + sid, e);
        } catch (AdminException e) {
            logger.error("Unable to remove the server " + sid, e);
            throw new JORAMServerException("Unable to remove the server " + sid, e);
        }
    }

    /**
     * @param config
     * @throws JORAMServerException
     */
    public static void addDomain(final Config config) throws JORAMServerException {

        if (config.getDomainPort() == null) {
            logger.error("Unable to add a domain without its domain name");
            throw new JORAMServerException(
            "Unable to add a domain without its domain name");
        }
        try {
            List<String> domains = Arrays.asList(AdminModule.getDomainNames(config.getServerId()));
            logger.debug("Server {0} belongs to the domains {1}", config.getServerId(), domains);
            if(!domains.contains(config.getDomainName())) {
                AdminModule.addDomain(config.getDomainName(),
                        config.getServerId(), config.getDomainPort());
                logger.debug("New domain added {0} to the server {1}", config.getDomainPort(), config.getServerId());
            }
        } catch (ConnectException e) {
            logger.error("Unable to add the domain " + config.getDomainName(), e);
            throw new JORAMServerException("Unable to add the domain " + config.getDomainName(), e);
        } catch (AdminException e) {
            logger.error("Unable to add the domain " + config.getDomainName(), e);
            throw new JORAMServerException("Unable to add the domain " + config.getDomainName(), e);
        }
    }

}
